/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s): 
 * Richard Kunze
 * Ole Arndt
 * David Li
 * Jacob Kjome
 *
 * $Id: DocumentLoaderImpl.java,v 1.19 2008-01-29 06:02:35 jkjome Exp $
 */

package org.enhydra.xml.xmlc.deferredparsing;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.lang.reflect.Field;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.Iterator;
import java.util.Map;
import java.util.WeakHashMap;

import org.enhydra.xml.io.ErrorReporter;
import org.enhydra.xml.lazydom.LazyDocument;
import org.enhydra.xml.lazydom.TemplateDOM;
import org.enhydra.xml.xmlc.XMLCError;
import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.xmlc.XMLCLogger;
import org.enhydra.xml.xmlc.XMLCRuntimeException;
import org.enhydra.xml.xmlc.XMLObject;
import org.enhydra.xml.xmlc.commands.xmlc.XMLCOptionsParser;
import org.enhydra.xml.xmlc.compiler.EditDOM;
import org.enhydra.xml.xmlc.compiler.Parse;
import org.enhydra.xml.xmlc.dom.XMLCDocument;
import org.enhydra.xml.xmlc.metadata.DocumentClass;
import org.enhydra.xml.xmlc.metadata.MetaData;
import org.w3c.dom.Document;

/**
 * Default implementation of the {@link DocumentLoader} interface.
 *
 * <p>This implementation maintains a global cache of loaded DOM
 * objects. If a DOM object is found in cache and the cached copy is
 * up to date with regard to the source document and XMLC metadata
 * object, a copy of the cached object is handed out. Otherwise, the
 * document is re-parsed from the source document, the DOM edits
 * specified in the document metadata are performed and the document
 * is stored in cache as a template. Note that the cache may be a
 * {@link WeakHashMap} or a map with similar semantics that frees up
 * entries if memory gets low.</p>
 *
 * <p>This class can be used as the basis for specialized
 * document loaders.</p>
 */
public class DocumentLoaderImpl implements DocumentLoader {

    /** The subfix of meta file names */
    final private static String META_FILE_EXT = ".xmlc";

    /** The associated factory */
    private XMLCDeferredParsingFactory factory;

    /** Helper class for storing cached DOMs */
    protected static class CacheEntry {
        volatile long timestamp = -1;
        volatile Document document = null;
        volatile URL src = null;
        volatile boolean ready = false;
    }

    /** The map to use for caching parsed template DOMs */
    private Map templateCache;

    /**
     * Default constructor. Uses a {@link Cache} as template
     * DOM cache.
     * @see #DocumentLoaderImpl(Map)
     */
    public DocumentLoaderImpl() {
        this(null);
    }

    /**
     * Constructor. Uses the specified map as template cache.
     * @param cache the cache to use.
     */
    public DocumentLoaderImpl(Map cache) {
        if (cache == null) {
            templateCache = new Cache();
        } else {
            templateCache = cache;
        }
    }

   /**
    * Generates a cache key for the given document class.  May
    * be overridden by subclasses to enhance caching behavior.
    * 
    * @param docClass the class to generate the cache key for
    * @return the key used to identify this class in the internal parsed
    *         document cache
    */
    protected String makeCacheKey(Class docClass) {
        return docClass.getName();
    }

    /**
     * Get the cache entry for a document, initialize one of not
     * already exists
     * 
     * Note that the cache entry does not necessarily hold a template
     * - it may need to be initialized first. Subclasses overriding
     * this method must make sure that new cache entries are
     * constructed and put in the cache in a thread safe manner.
     *
     * @param docClass the class to get the cache entry for.
     * @return a cache entry. The returned object is suitable as mutex
     * for document intialization.
     */
    protected CacheEntry getCacheEntry(Class docClass) {
        String name = makeCacheKey(docClass);

        /* XXX: This assumes that concurrent unsynchronized accesses
         * to templateCache.get() with different keys are not
         * disturbed by the code in the synchronized block, and that
         * concurrent accesses with the same key return either null or
         * the entry that has been put into the map inside the
         * synchronized block. 
         *
         * The synchronization is broken down to two step when the
         * entry does not exist in the cache. First, lock the template
         * cache and create a empty entry and template cache is
         * unlocked.  Then the new empty entry is locked and
         * updated. Breaking down to two steps would avoid the entry
         * creating 
         */
        CacheEntry entry = (CacheEntry) templateCache.get(name);
        if (entry == null) {
            synchronized (templateCache) {
                entry = (CacheEntry) templateCache.get(name);
                if (entry == null) {

                    if (getLogger().debugEnabled()) {
                        getLogger().logDebug(">>>Creating cache entry for "
                                             + docClass.getName());
                    }

                    entry = new CacheEntry();
                    templateCache.put(name, entry);
                }
            }
        }

        if (!entry.ready) {
            synchronized (entry) {
                if (!entry.ready) {
                    entry.src = findSourceUrl(docClass);
                    entry.document = parseDocument(docClass, entry.src);
                    entry.timestamp = getCurrentTimestamp(docClass, entry.src);
                    entry.ready = true;
                }
            }
        }

        return entry;
    }

    /**
     * Check if the DOM instance is out of date
     */
    private CacheEntry checkCacheEntry(Class docClass, CacheEntry entry) {

        // Check if the document is out of date, and reload it if yes.
        long ts = getCurrentTimestamp(docClass, entry.src);

        if (getLogger().debugEnabled()) {
            getLogger().logDebug(">>>Checking time stamp ts="
                                 + new Date(ts));
            getLogger().logDebug(">>>       entry.timestamp="
                                 + new Date(entry.timestamp));
        }

        if (ts == -1 || ts > entry.timestamp) {
            synchronized (entry) {
                ts = getCurrentTimestamp(docClass, entry.src);
                // Check if some other thread did reparse the document
                // while this one has been waiting for the lock...
                if (ts == -1 || ts > entry.timestamp) {
                    entry.document = parseDocument(docClass, entry.src);
                    entry.timestamp = getCurrentTimestamp(docClass, entry.src);
                } else {
                    if (getLogger().debugEnabled()) {
                        getLogger().logDebug(">>>DOM for "
                                             + docClass.getName()
                                             + " has been updated by another thread.");
                    }
                }
            }
        }

        return entry;
    }

    /** 
     * Get a DOM instance representing the specified document.
     * @param docClass the XMLC generated class to get the DOM for.
     * @return the DOM representing objects of <code>docClass</code>
     */
    public Document getDocument(Class docClass) throws XMLCRuntimeException {
        if (getLogger().debugEnabled()) {
            getLogger().logDebug(">>>DOM instance requested for class "
                                 + docClass.getName());
        }

        CacheEntry entry = getCacheEntry(docClass); // get the cached entry

        entry = checkCacheEntry(docClass, entry);
        // make sure cache entry is up to date

        if (getLogger().debugEnabled()) {
            getLogger().logDebug(">>>Document class "
                                 + entry.document.getClass().getName());
        }

        return (Document) entry.document.cloneNode(true);
    }

    /**
     * Get one of the compiled-in constants in a XMLC-generated class (or
     * subclass).
     */
    public static Object getClassConstant(Class xmlcBasedClass,
                                          String constName) {
        // Use generated constant to find the class.
        try {
            Field field = xmlcBasedClass.getField(constName);
            return field.get(null);
        } catch (NoSuchFieldException except) {
            throw new XMLCRuntimeException("Couldn't find class constant \""
                                           + constName
                                           + "\" in class \""
                                           + xmlcBasedClass.getName()
                                           + "\", maybe be a XMLC generated class or descendent",
                                           except);
        } catch (IllegalAccessException except) {
            throw new XMLCRuntimeException(except);
        }
    }

    /** 
     * Get the name of the source file associated with <code>docClass</code> 
     */
    protected String getSourceFileName(Class docClass)
        throws XMLCRuntimeException {
        String srcPath =
            (String) getClassConstant(docClass,
                                      XMLObject.XMLC_SOURCE_FILE_FIELD_NAME);
        if (srcPath == null) {
            throw new XMLCRuntimeException(XMLObject.XMLC_SOURCE_FILE_FIELD_NAME
                                           + " is null in class "
                                           + docClass.getName());
        }
        if (srcPath.startsWith("/")) {
            srcPath = srcPath.substring(1);
        }
        return srcPath;
    }

    /**
     * Get the document URL for a document on the classpath
     * 
     * @param path   the fully qualified path to the source file
     * @param loader the class loader used to locate the source file
     */
    private URL getDocURLFromClasspath(String path, ClassLoader loader) {
        URL srcURL = loader.getResource(path);

        if ((srcURL != null) && getLogger().debugEnabled()) {
            getLogger().logDebug(">>>Get document '"
                                 + srcURL
                                 + "' from classpath");
        }
        return srcURL;
    }

    /**
     * Get the document URL for a document on the resource dir path
     *
     * @param path the relative path to the source file
     */
    private URL getDocURLFromResourceDir(String path) {
        File srcFile = null;

        for (Iterator iter = getFactory().getResourceDirList().iterator(); iter.hasNext();) {
            File dir = (File) iter.next();
            srcFile = new File(dir, path);
            /* checking under dir using long path */
            if (srcFile.exists()) {
                break;
            }
            srcFile = null;
        }

        if (srcFile != null) {
            try {
                URL srcURL =  srcFile.toURL(); //TODO - replicate file.toURI().toURL() (for proper escaping), only jdk1.3 compatible
                if (getLogger().debugEnabled()) {
                    getLogger().logDebug(">>>Get document '"
                                         + srcURL
                                         + "' from resource dir");
                }
                return srcURL;                
            } catch (MalformedURLException e) {
                throw new XMLCError("Cannot construct URL for file " + srcFile, e);
            }
        }
        return null;        
    }

    /**
     * Look up each path from the resource dir and fall back to the classpath if
     * it isn't found.  If it isn't found in the classpath, repeat with the other
     * paths.  If no path is found in any resource dir or the classloader, return
     * null.
     * 
     * @param paths  the relative paths to the source files (lone filename or
     *               directory path to file with no leading slashes,
     *               eg... "some/path/to/File.ext") 
     * @param loader the class loader used to locate a source file on the classpath
     * @return the url of the path that is found or null if no path is found
     */
    private URL getDocURLFromResourceDirOrClasspath(String[] paths, ClassLoader loader) {
        URL url = null;
        for (int i = 0; i < paths.length; i++) {
            //search in resource dirs first so one can copy files to the classloader
            //for fallback, but still be able to override files located in the
            //classloader by putting files in a resource path. Makes for simpler
            //development and more flexible deployment.
            url = getDocURLFromResourceDir(paths[i]);
            if (url == null) {
                url = getDocURLFromClasspath(paths[i], loader);
            }

            if (url != null) {
                break;
            }
        }
        return url;
    }

    /**
     * Get possible file paths by subtracting each respective package prefix
     * from the path prefix.  The original path is included last in the returned
     * value.
     */
    private String[] getPathsFromPackagePrefixes(String path) {
        ArrayList list = new ArrayList();

        for (Iterator iter = getFactory().getPackagePrefixList().iterator(); iter.hasNext();) {
            String prefix = (String) iter.next();

            if (path.startsWith(prefix)) {
                list.add(path.substring(prefix.length() + 1));
                break;
            }
        }
        //Add the full path last so that package prefix paths can be evaluated first
        list.add(path);

        return (String[]) list.toArray(new String[0]);
    }

    /**
     * Find the URL of the source document associated with <code>docClass</code>.
     * If the source file cannot be found, an exception is thrown.
     * Resource dirs are searched first, then the classloader.
     * Package prefix paths are taken into account.
     * 
     * @param docClass the class for which to load the source doc
     * @throws XMLCRuntimeException if the source doc cannot be found
     */
    protected URL findSourceUrl(Class docClass) {
        if (getLogger().debugEnabled()) {
            getLogger().logDebug(">>>Loading source doc for "
                                 + docClass.getName());
        }

        String srcPath = getSourceFileName(docClass);
        String[] paths = getPathsFromPackagePrefixes(srcPath);
        URL srcURL = getDocURLFromResourceDirOrClasspath(paths, docClass.getClassLoader());

        if (srcURL == null) {
            throw new XMLCRuntimeException("Source path(s) '"
                                           + Arrays.asList(paths)
                                           + "' not found for "
                                           + docClass.getName());
        }
        return srcURL;
    }

    /**
     * Get the meta-data file path.  This is the interface name for the
     * class with the .xmlc extension.
     */
    protected String metaDataFileName(Class docClass) {
        //Convert class to file path, drop Impl and add extension.
        String fileName = docClass.getName().replace('.', '/');
        if (fileName.endsWith(DocumentClass.IMPLEMENTATION_SUFFIX)) {
            fileName =
                fileName.substring(0,
                                   fileName.length()
                                   - DocumentClass.IMPLEMENTATION_SUFFIX.length());
        }
        return fileName + META_FILE_EXT;
    }

    /**
     * Get the meta-data file paths.  Provides an array containing paths
     * representing the FQCN of the metadata file, plus the .xmlc extension, as
     * well as every combination of paths corresponding to configured package
     * prefixes.  Paths represent metadata files for both interface and "Impl"
     * classes.  This covers all cases of .xmlc filename generation.
     */
    private String[] getMetaDataFileNames(Class docClass) {
        String defaultFileName = metaDataFileName(docClass); 

        String[] A = getPathsFromPackagePrefixes(defaultFileName);
        String[] B = getPathsFromPackagePrefixes(defaultFileName.substring(0, defaultFileName.lastIndexOf(META_FILE_EXT)) + DocumentClass.IMPLEMENTATION_SUFFIX + META_FILE_EXT);
        String[] C = new String[A.length + B.length];
        System.arraycopy(A, 0, C, 0, A.length);
        System.arraycopy(B, 0, C, A.length, B.length);
        //perform sort to get a consistent lookup order when dealing with
        //package prefix paths and interface-vs-Impl extensions in corresponding
        //paths
        Arrays.sort(C);
        return C;
    }

    /**
     * Load the XMLC metadata file.  If the class-specific metadata file cannot
     * be found, the default metadata file searched for.  If the default
     * metadata file cannot be found (or has not been defined), an exception
     * is thrown.  Resource dirs are searched first, then the classloader.
     * Package prefix paths are taken into account.
     * 
     * @param docClass the class for which to load the metadata
     * @param errorReporter the error handler to use
     * @throws XMLCRuntimeException if the metadata cannot be found
     * @throws XMLCException if the found metadata cannot be parsed
     */
    protected MetaData loadMetaData(Class docClass,
                                    ErrorReporter errorReporter)
        throws XMLCException {
        if (getLogger().debugEnabled()) {
            getLogger().logDebug(">>>Loading metadata for "
                                 + docClass.getName());
        }

        String[] paths = getMetaDataFileNames(docClass);
        URL metaDataURL = getDocURLFromResourceDirOrClasspath(paths, docClass.getClassLoader());

        if (metaDataURL == null) {
            String defaultMetaFile = factory.getDefaultMetaDataFile();
            if (defaultMetaFile != null) {
                if (defaultMetaFile.indexOf('/') == -1) {
                    //convert dots to slashes while preserving the file extension
                    int lastIndex = defaultMetaFile.lastIndexOf(".");
                    if (lastIndex != -1) {
                        String defaultPath = defaultMetaFile.substring(0, lastIndex);
                        defaultPath = defaultPath.replace('.', '/');
                        defaultMetaFile = defaultPath + defaultMetaFile.substring(lastIndex);
                    }
                }

                String[] defaultPaths = getPathsFromPackagePrefixes(defaultMetaFile);
                metaDataURL = getDocURLFromResourceDirOrClasspath(defaultPaths, docClass.getClassLoader());
            }
        }

        if (metaDataURL == null) {
            throw new XMLCRuntimeException("Metadata path(s) '"
                                           + Arrays.asList(paths)
                                           + "' not found for "
                                           + docClass.getName());
        }

        try {
            //NOTE: Use of "-version" gets around an implementation detail of
            //XMLCOptionsParser.parse() where no filename is required if
            //"-version" is provided.  Might want to revisit this later to make
            //it less hackish.
            return new XMLCOptionsParser().parse(new String[]{"-version", metaDataURL.toString()}, errorReporter, docClass.getClassLoader());
        } catch (IOException ioe) {
            throw new XMLCException(ioe);
        }
    }

    /**
     * Get the current timestamp of the *ML template for docClass.
     * @param docClass the class to get the information for
     * @param src the source URL for this class, as returned by
     * {@link #findSourceUrl(Class)}
     * @return the timestamp, or -1 to force a reload.
     */
    protected long getCurrentTimestamp(Class docClass, URL src) {
        if ("file".equals(src.getProtocol())) {
            return new File(src.getPath()).lastModified();
        }

        URLConnection urlc = null;
        try {
            urlc = src.openConnection();
            urlc.setUseCaches(false);
            return urlc.getLastModified();
        } catch (IOException e) {
            getLogger().logInfo("Cannot read last modified time for " + src, e);
            return -1;
        } finally {
            if (urlc != null) {
                try {
                    urlc.getInputStream().close();
                } catch (Exception e) {
                    //squash
                }
            }
        }
    }

    /** Create a new template DOM for docClass
     * @param docClass the class to load the DOM for
     * @param src the source file for this class, as returned by
     * {@link #findSourceFile}
     */
    protected Document parseDocument(Class docClass, URL src) {
        ErrorReporter errorReporter = new ErrorReporter();
        try {
            if (getLogger().debugEnabled()) {
                getLogger().logDebug(">>>Parsing DOM for "
                                     + docClass.getName()
                                     + " from source URL "
                                     + src.toString());
            }

            MetaData metaData = loadMetaData(docClass, errorReporter);
            metaData.getInputDocument().setUrl(src.toString());

            Parse parser = createParser(docClass, errorReporter, null);
            XMLCDocument xmlcDoc = parser.parse(metaData);

            EditDOM domEditor = createDOMEditor(docClass, metaData);
            domEditor.edit(xmlcDoc);

            if (xmlcDoc.getDocument() instanceof LazyDocument) {
                LazyDocument lazyDoc = (LazyDocument)xmlcDoc.getDocument();
                //TODO - Do we really need a clone here, or just a way to set the templateDOM and reset flags to use it properly?
                //The nodes are the same.  The problem is that they end up not being owned by the new document, but by the
                //delegate document inside the TemplateDOM.  This requires extended documents to override cloneAsTemplateDocument()
                //as well as LazyDocument overriding getIdentifier() because it compares the element's parent to "this" which
                //fails because of the delegation.  However, we might need to override anyway to avoid unnecessary expansion of nodes.
                xmlcDoc.setDocument(lazyDoc.cloneAsTemplateDocument(new TemplateDOM(lazyDoc)));
            }

            return xmlcDoc.getDocument();
        } catch (Exception e) {
            XMLCRuntimeException except;
            if (e instanceof XMLCRuntimeException) {
                except = (XMLCRuntimeException) e;
            } else {
                except = new XMLCRuntimeException(e);
            }
            throw except;
        }
    }

    /** Create the parser to use. Override this if you need a special
     * version */
    protected Parse createParser(Class docClass,
                                 ErrorReporter errorReporter,
                                 PrintWriter verboseOut) {
        return new Parse(errorReporter, verboseOut);
    }

    /** Create the DOM editor to use. Override this if you need a special
     * version */
    protected EditDOM createDOMEditor(Class docClass, MetaData metaData) {
        return new EditDOM(metaData);
    }

    /**
     * Bind to a factory.
     *
     * @param factory Factory that is creating the object.
     */
    public void init(XMLCDeferredParsingFactory factory) {
        this.factory = factory;
    }

    /** Get the associated factory */
    protected XMLCDeferredParsingFactory getFactory() {
        return factory;
    }

    /** Get the associated logger */
    protected XMLCLogger getLogger() {
        return factory.getLogger();
    }
}
